@@ -1,6 +1,7 @@
 """Web mercator XYZ tile utilities"""
 
 from collections import namedtuple
+from functools import reduce
 import math
 import sys
 import warnings
@@ -12,11 +13,22 @@ if sys.version_info < (3,):
         UserWarning,
     )
     from collections import Sequence
+
+    def lru_cache(maxsize=None):
+        """Does nothing. We do not cache for Python < 3."""
+
+        def fake_decorator(func):
+            return func
+
+        return fake_decorator
+
+
 else:
     from collections.abc import Sequence
+    from functools import lru_cache
 
 
-__version__ = "1.1.6"
+__version__ = "1.2.1"
 
 __all__ = [
     "Bbox",
@@ -28,6 +40,7 @@ __all__ = [
     "children",
     "feature",
     "lnglat",
+    "neighbors",
     "parent",
     "quadkey",
     "quadkey_to_tile",
@@ -36,8 +49,10 @@ __all__ = [
     "tiles",
     "ul",
     "xy_bounds",
+    "minmax",
 ]
 
+
 R2D = 180 / math.pi
 RE = 6378137.0
 CE = 2 * math.pi * RE
@@ -45,14 +60,25 @@ EPSILON = 1e-14
 LL_EPSILON = 1e-11
 
 
-Tile = namedtuple("Tile", ["x", "y", "z"])
-"""An XYZ web mercator tile
+class Tile(namedtuple("Tile", ["x", "y", "z"])):
+    """An XYZ web mercator tile
 
-Attributes
-----------
-x, y, z : int
-    x and y indexes of the tile and zoom level z.
-"""
+    Attributes
+    ----------
+    x, y, z : int
+        x and y indexes of the tile and zoom level z.
+
+    """
+
+    def __new__(cls, x, y, z):
+        """A new instance"""
+        lo, hi = minmax(z)
+        if not lo <= x <= hi or not lo <= y <= hi:
+            warnings.warn(
+                "Mercantile 2.0 will require tile x and y to be within the range (0, 2 ** zoom)",
+                FutureWarning,
+            )
+        return tuple.__new__(cls, [x, y, z])
 
 
 LngLat = namedtuple("LngLat", ["lng", "lat"])
@@ -176,24 +202,28 @@ def bounds(*tile):
 
     Parameters
     ----------
-    tile : Tile or sequence of int
-        May be be either an instance of Tile or 3 ints, X, Y, Z.
+    tile : Tile or tuple
+        May be be either an instance of Tile or 3 ints (X, Y, Z).
 
     Returns
     -------
-    LngLatBBox
-
-    Notes
-    -----
-    Epsilon is subtracted from the right limit and added to the bottom
-    limit.
+    LngLatBbox
 
     """
     tile = _parse_tile_arg(*tile)
     xtile, ytile, zoom = tile
-    a = ul(xtile, ytile, zoom)
-    b = ul(xtile + 1, ytile + 1, zoom)
-    return LngLatBbox(a[0], b[1], b[0], a[1])
+
+    Z2 = math.pow(2, zoom)
+
+    ul_lon_deg = xtile / Z2 * 360.0 - 180.0
+    ul_lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * ytile / Z2)))
+    ul_lat_deg = math.degrees(ul_lat_rad)
+
+    lr_lon_deg = (xtile + 1) / Z2 * 360.0 - 180.0
+    lr_lat_rad = math.atan(math.sinh(math.pi * (1 - 2 * (ytile + 1) / Z2)))
+    lr_lat_deg = math.degrees(lr_lat_rad)
+
+    return LngLatBbox(ul_lon_deg, lr_lat_deg, lr_lon_deg, ul_lat_deg)
 
 
 def truncate_lnglat(lng, lat):
@@ -264,6 +294,59 @@ def lnglat(x, y, truncate=False):
     return LngLat(lng, lat)
 
 
+def neighbors(*tile, **kwargs):
+    """The neighbors of a tile
+
+    The neighbors function makes no guarantees regarding neighbor tile
+    ordering.
+
+    The neighbors function returns up to eight neighboring tiles, where
+    tiles will be omitted when they are not valid e.g. Tile(-1, -1, z).
+
+    Parameters
+    ----------
+    tile : Tile or sequence of int
+        May be be either an instance of Tile or 3 ints, X, Y, Z.
+
+    Returns
+    -------
+    list
+
+    Examples
+    --------
+    >>> neighbors(Tile(486, 332, 10))
+    [Tile(x=485, y=331, z=10), Tile(x=485, y=332, z=10), Tile(x=485, y=333, z=10), Tile(x=486, y=331, z=10), Tile(x=486, y=333, z=10), Tile(x=487, y=331, z=10), Tile(x=487, y=332, z=10), Tile(x=487, y=333, z=10)]
+
+    """
+    xtile, ytile, ztile = _parse_tile_arg(*tile)
+
+    tiles = []
+
+    lo, hi = minmax(ztile)
+
+    for i in [-1, 0, 1]:
+        for j in [-1, 0, 1]:
+            if i == 0 and j == 0:
+                continue
+            elif xtile + i < 0 or ytile + j < 0:
+                continue
+            elif xtile + i > hi or ytile + j > hi:
+                continue
+            tiles.append(Tile(x=xtile + i, y=ytile + j, z=ztile))
+
+    # Make sure to not generate invalid tiles for valid input
+    # https://github.com/mapbox/mercantile/issues/122
+    def valid(tile):
+        validx = 0 <= tile.x <= 2 ** tile.z - 1
+        validy = 0 <= tile.y <= 2 ** tile.z - 1
+        validz = 0 <= tile.z
+        return validx and validy and validz
+
+    tiles = [t for t in tiles if valid(t)]
+
+    return tiles
+
+
 def xy_bounds(*tile):
     """Get the web mercator bounding box of a tile
 
@@ -484,25 +567,26 @@ def parent(*tile, **kwargs):
 
     Examples
     --------
-
     >>> parent(Tile(0, 0, 2))
     Tile(x=0, y=0, z=1)
-
     >>> parent(Tile(0, 0, 2), zoom=0)
     Tile(x=0, y=0, z=0)
 
     """
     tile = _parse_tile_arg(*tile)
+    x, y, z = tile
+
+    if z == 0:
+        return None
 
     # zoom is a keyword-only argument.
     zoom = kwargs.get("zoom", None)
 
-    if zoom is not None and (tile[2] < zoom or zoom != int(zoom)):
+    if zoom is not None and (z <= zoom or zoom != int(zoom)):
         raise InvalidZoomError(
             "zoom must be an integer and less than that of the input tile"
         )
 
-    x, y, z = tile
     if x != int(x) or y != int(y) or z != int(z):
         raise ParentTileError("the parent of a non-integer tile is undefined")
 
@@ -533,19 +617,24 @@ def children(*tile, **kwargs):
     tile : Tile or sequence of int
         May be be either an instance of Tile or 3 ints, X, Y, Z.
     zoom : int, optional
-        Returns all children at zoom *zoom*, in depth-first clockwise winding order.
-        If unspecified, returns the immediate (i.e. zoom + 1) children of the tile.
+        Returns all children at zoom *zoom*, in depth-first clockwise
+        winding order.  If unspecified, returns the immediate (i.e. zoom
+        + 1) children of the tile.
 
     Returns
     -------
     list
 
+    Raises
+    ------
+    InvalidZoomError
+        If the zoom level is not an integer greater than the zoom level
+        of the input tile.
+
     Examples
     --------
-
     >>> children(Tile(0, 0, 0))
     [Tile(x=0, y=0, z=1), Tile(x=0, y=1, z=1), Tile(x=1, y=0, z=1), Tile(x=1, y=1, z=1)]
-
     >>> children(Tile(0, 0, 0), zoom=2)
     [Tile(x=0, y=0, z=2), Tile(x=0, y=1, z=2), Tile(x=0, y=2, z=2), Tile(x=0, y=3, z=2), ...]
 
@@ -565,6 +654,7 @@ def children(*tile, **kwargs):
     target_zoom = zoom if zoom is not None else ztile + 1
 
     tiles = [tile]
+
     while tiles[0][2] < target_zoom:
         xtile, ytile, ztile = tiles.pop(0)
         tiles += [
@@ -573,6 +663,7 @@ def children(*tile, **kwargs):
             Tile(xtile * 2 + 1, ytile * 2 + 1, ztile + 1),
             Tile(xtile * 2, ytile * 2 + 1, ztile + 1),
         ]
+
     return tiles
 
 
@@ -618,7 +709,7 @@ def simplify(tiles):
     for tile in sorted(tiles, key=operator.itemgetter(2)):
         x, y, z = tile
         is_new_tile = True
-        for supertile in (parent(tile, zoom=i) for i in range(z + 1)):
+        for supertile in (parent(tile, zoom=i) for i in range(z)):
             if supertile in root_set:
                 is_new_tile = False
                 continue
@@ -761,3 +852,96 @@ def feature(
         feat["id"] = fid
 
     return feat
+
+
+def _coords(obj):
+    """All coordinate tuples from a geometry or feature or collection
+
+    Yields
+    ------
+    lng : float
+        Longitude
+    lat : float
+        Latitude
+
+    """
+    if isinstance(obj, (tuple, list)):
+        coordinates = obj
+    elif "features" in obj:
+        coordinates = [feat["geometry"]["coordinates"] for feat in obj["features"]]
+    elif "geometry" in obj:
+        coordinates = obj["geometry"]["coordinates"]
+    else:
+        coordinates = obj.get("coordinates", obj)
+
+    for e in coordinates:
+        if isinstance(e, (float, int)):
+            yield tuple(coordinates)
+            break
+        else:
+            for f in _coords(e):
+                yield f[:2]
+
+
+def geojson_bounds(obj):
+    """Returns the bounding box of a GeoJSON object
+
+    Parameters
+    ----------
+    obj : mapping
+        A GeoJSON geometry, feature, or feature collection.
+
+    Returns
+    -------
+    LngLatBbox
+
+    """
+
+    def func(bbox, coords):
+        w, s, e, n = bbox
+        lng, lat = coords
+        return min(w, lng), min(s, lat), max(e, lng), max(n, lat)
+
+    w, s, e, n = reduce(func, _coords(obj), (180.0, 90.0, -180.0, -90.0))
+    return LngLatBbox(w, s, e, n)
+
+
+@lru_cache(maxsize=28)
+def minmax(zoom):
+    """Minimum and maximum tile coordinates for a zoom level
+
+    Parameters
+    ----------
+    zoom : int
+        The web mercator zoom level.
+
+    Returns
+    -------
+    minimum : int
+        Minimum tile coordinate (note: always 0).
+    maximum : int
+        Maximum tile coordinate (2 ** zoom - 1).
+
+    Raises
+    ------
+    InvalidZoomError
+        If zoom level is not a positive integer.
+
+    Examples
+    --------
+    >>> minmax(1)
+    (0, 1)
+    >>> minmax(-1)
+    Traceback (most recent call last):
+    ...
+    InvalidZoomError: zoom must be a positive integer
+
+    """
+
+    try:
+        if int(zoom) != zoom or zoom < 0:
+            raise InvalidZoomError("zoom must be a positive integer")
+    except ValueError:
+        raise InvalidZoomError("zoom must be a positive integer")
+
+    return (0, 2 ** zoom - 1)
